package zander.library;

import java.io.File;
import java.io.IOException;
import java.net.URI;
import java.util.function.Consumer;

import javax.swing.SwingUtilities;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import zander.library.Manifest.ManifestDifference;
import zander.library.Manifest.ManifestParseException;
import zander.library.secrets.Secrets;

public class Library {
    private static final Logger LOGGER = LoggerFactory.getLogger(Library.class);

    public static final class SyncStatus {
        private int total = -1;
        private int current;
        private boolean started;
        private Throwable error;
        private boolean canceled;
        private String note;
        private boolean done;

        private Consumer<SyncStatus> updateAction;

        public SyncStatus(Consumer<SyncStatus> updateAction) {
            this.updateAction = updateAction;
        }

        private synchronized void finish() {
            done = true;
            runUpdateAction();
        }

        public synchronized boolean isDone() {
            return done;
        }

        public synchronized String getNote() {
            return note;
        }

        private synchronized void setNote(String note) {
            LOGGER.info(note);
            this.note = note;
            runUpdateAction();
        }

        private synchronized void setTotal(int total) {
            this.total = total;
            runUpdateAction();
        }

        public synchronized boolean isStarted() {
            return started;
        }

        public synchronized Consumer<SyncStatus> getUpdateAction() {
            return updateAction;
        }

        public synchronized Throwable getError() {
            return error;
        }

        public synchronized boolean isFailed() {
            return error != null;
        }

        public synchronized void setUpdateAction(Consumer<SyncStatus> updateAction) {
            this.updateAction = updateAction;
        }

        public synchronized int getTotal() {
            return total;
        }

        public synchronized int getCurrent() {
            return current;
        }

        public synchronized float getPrecent() {
            return (float) current / total;
        }

        public synchronized void cancel() {
            canceled = true;
        }

        public synchronized boolean isCanceled() {
            return canceled;
        }

        private synchronized void incrementCurrent() {
            ++this.current;
            runUpdateAction();
        }

        private synchronized void runUpdateAction() {
            if (updateAction != null) {
                SwingUtilities.invokeLater(() -> updateAction.accept(this));
            }
        }

    }

    public static final String CONFIG_FILE_PATH = "config.json";
    public static final String ALBUMS_FILE_PATH = "albums";
    public static final String THUMBNAILS_FILE_PATH = "thumbnails";
    public static final String ALBUM_CONFIG_FILE_NAME = "data.json";

    private LocalCache localCache;
    private RemoteStore remoteStore;

    public Library(File local, URI remote, boolean passiveMode, Secrets secrets) throws ManifestParseException, IOException {
        remoteStore = new RemoteStore(remote, secrets);
        remoteStore.setPassiveMode(passiveMode);
        localCache = new LocalCache(local);
    }

    private void syncLocalToRemote(SyncStatus status) throws IOException, ManifestParseException {
        if (status != null && !status.isStarted()) {
            Thread t = new Thread(() -> {
                    status.started = true;
                    status.setNote("Fetching manifest...");
                    try {
                        syncLocalToRemote(status);
                    } catch (Throwable e) {
                        LOGGER.error("Failed to sync local to remote!", e);
                        status.error = e;
                        status.runUpdateAction();
                        return;
                    }
            });
            t.start();
            return;
        }
        LOGGER.info("Syncing local cahce to remote store...");
        remoteStore.doFtpActions(() -> {
            status.setNote("Retreving remote manifest");
            remoteStore.fetchManifest();
            Manifest rm = remoteStore.getManifest();
            status.setNote("Reading remote manifest");
            Manifest lm = localCache.getManifest();
            status.setNote("Calculating differences...");
            ManifestDifference diff = lm.compareTo(rm);
            status.setTotal(diff.newFiles.size() + diff.removedFiles.size());
            if (status != null) {
                if (status.isCanceled()) {
                    LOGGER.info("Sync canceled!");
                    return;
                }
            }
            status.setNote("Deleting old files...");
            for (String f : diff.removedFiles) {
                localCache.deleteFile(f);
                status.incrementCurrent();
                status.setNote("Deleted: " + f);
                if (status != null) {
                    if (status.isCanceled()) {
                        LOGGER.info("Sync canceled!");
                        return;
                    }
                }
            }
            status.setNote("Retrieving new and updated files...");
            for (String f : diff.newFiles) {
                status.setNote("Retrieving '" + f + "'..");
                byte[] data = remoteStore.retrieveFile(f);
                localCache.addFile(f, data);
                status.setNote("Retrived: '" + f + "' (" + data.length + " B)");
                status.incrementCurrent();
                if (status != null) {
                    if (status.isCanceled()) {
                        LOGGER.info("Sync canceled!");
                        return;
                    }
                }
            }
        });
        if (!status.isCanceled()) {
            status.setNote("Done!");
            status.finish();
        }
    }

    public void syncLocalToRemote() throws IOException, ManifestParseException {
        syncLocalToRemote(null);
    }

    public SyncStatus syncLocalToRemoteAsync(Consumer<SyncStatus> updateAction) {
        SyncStatus s = new SyncStatus(updateAction);
        try {
            syncLocalToRemote(s);
        } catch (ManifestParseException | IOException e) {
            // ignore
        }
        return s;
    }

    public void clearLocalCache() throws IOException {
        localCache.clear();
    }

    public LocalCache getLocalCache() {
        return localCache;
    }

    public RemoteStore getRemoteStore() {
        return remoteStore;
    }

    public void addFile(String name, byte[] data) throws IOException {
        localCache.addFile(name, data);
        remoteStore.addFile(name, data);
    }

    public void addBinaryFile(String name, byte[] data) throws IOException {
        localCache.addFile(name, data);
        remoteStore.addBinaryFile(name, data);
    }

    public void addAsciiFile(String name, String data) throws IOException {
        localCache.addFile(name, data);
        remoteStore.addAsciiFile(name, data);
    }

    public void deleteFile(String name) throws IOException {
        localCache.deleteFile(name);
        remoteStore.deleteFile(name);
    }

    public void moveFile(String name, String target) throws IOException {
        localCache.moveFile(name, target);
        remoteStore.moveFile(name, target);
    }
}
