package zander.library.media;

import java.awt.Component;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;

import javax.swing.SwingUtilities;

import org.apache.tika.Tika;

import zander.library.FileUtils;
import zander.ui.ErrorDialog;
import zander.ui.UIUtils;

public class UnknownMediaData implements MediaData {
    private static final Tika TIKA = new Tika();

    private final SerializableImageContainer errorImage;
    private final ImageMediaComponent viewer;
    private final String type;

    private File sourceFile;

    public UnknownMediaData(File file) throws IOException {
        this(file, readFileBytes(file));
    }

    public UnknownMediaData(byte[] data) throws IOException {
        this(FileUtils.createTempFile(data, true), data);
    }

    public UnknownMediaData(File file, byte[] data) {
        errorImage = new SerializableImageContainer(getErrorIcon());
        viewer = new ImageMediaComponent(errorImage.getImage());
        type = TIKA.detect(data);
        sourceFile = file;
    }

    @Override
    public byte[] loadRawBytes() throws IOException {
        return readFileBytes(sourceFile);
    }

    @Override
    public String getType() {
        return type;
    }

    @Override
    public Component createThumbnailComponent() {
        ImageMediaComponent c = new ImageMediaComponent(errorImage.getImage());
        c.setPreferredSize(UIUtils.THUMBNAIL_SIZE);
        return c;
    }

    @Override
    public Component getViewerComponent() {
        return viewer;
    }

    @Override
    public File getSourceFile() {
        return sourceFile;
    }

    @Override
    public void setSourceFile(File file) {
        this.sourceFile = file;
    }

    @Override
    public boolean hasSourceFile() {
        return sourceFile != null;
    }

    private static byte[] readFileBytes(File file) throws IOException{
        try (FileInputStream in = new FileInputStream(file)) {
            return in.readAllBytes();
        }
    }

    private static BufferedImage getErrorIcon() {
        try {
            BufferedImage i = UIUtils.loadBufferedImage(ClassLoader.getSystemResource("missing-image-icon.svg"), -1, -1);
            return i;
        } catch (IOException e) {
            SwingUtilities.invokeLater(() -> {
                    ErrorDialog ed = new ErrorDialog("Internal Error", "A fatal internal error has occured!", e);
                    ed.setVisible(true);
                });
            System.err.println("Could not load missing image icon!");
            System.exit(1);
            return null;
        }
    }
}
