package zander.library.media;

import java.awt.Component;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;

import javax.swing.SwingUtilities;

import org.apache.tika.Tika;

import zander.library.FileUtils;
import zander.ui.ErrorDialog;
import zander.ui.ProgressTracker.ProgressManager;
import zander.ui.UIUtils;

public class VideoMediaData implements MediaData {
    private static final Tika TIKA = new Tika();
    private static final SerializableImageContainer ERROR_ICON = getErrorIcon();

    private SerializableImageContainer thumbnail;
    private final VideoMediaViewer viewer;
    private final String type;

    public VideoMediaData(File file, ProgressManager t) throws IOException {
        this(readAllBytes(file), t);
    }

    public VideoMediaData(byte[] data, ProgressManager t) throws IOException {
        this(FileUtils.createTempFile(data, true), data, t);
    }

    public VideoMediaData(File file, byte[] data, ProgressManager t) {
        try {
            BufferedImage i = getThumbnailFrame(data, t);
            if (i != null) {
                this.thumbnail = new SerializableImageContainer(i);
            } else {
                this.thumbnail = ERROR_ICON;
            }
        } catch (Throwable e) {
            e.printStackTrace();
            this.thumbnail = ERROR_ICON;
        }
        viewer = new VideoMediaViewer(file, this.thumbnail);
        type = TIKA.detect(data);
    }

    public SerializableImageContainer getThumbnailFrame() {
        return thumbnail;
    }

    @Override
    public byte[] loadRawBytes() throws IOException {
        return readAllBytes(getSourceFile());
    }

    @Override
    public String getType() {
        return type;
    }

    @Override
    public Component createThumbnailComponent() {
        ImageMediaComponent thumbnail = new ImageMediaComponent(this.thumbnail);
        thumbnail.setPreferredSize(UIUtils.THUMBNAIL_SIZE);
        return thumbnail;
    }

    @Override
    public Component getViewerComponent() {
        return viewer;
    }

    @Override
    public File getSourceFile() {
        return viewer.getSourceFile();
    }

    @Override
    public void setSourceFile(File file) {
        viewer.setSourceFile(file);
    }

    @Override
    public boolean hasSourceFile() {
        return viewer.getSourceFile() != null;
    }

    private static byte[] readAllBytes(File file) throws IOException {
        FileInputStream in = new FileInputStream(file);
        byte[] b = in.readAllBytes();
        in.close();
        return b;
    }

    private static BufferedImage getThumbnailFrame(byte[] data, ProgressManager t) throws IOException {
        BufferedImage img = FFmpeg.createThumbnail(data, t);
        return UIUtils.fitImageToBox(img, UIUtils.IMAGE_PREVEW_SIZE);
    }

    private static SerializableImageContainer getErrorIcon() {
        try {
            BufferedImage i = UIUtils.loadBufferedImage(ClassLoader.getSystemResource("video-icon.png"), -1, -1);
            return new SerializableImageContainer(i);
        } catch (IOException e) {
            SwingUtilities.invokeLater(() -> {
                    ErrorDialog ed = new ErrorDialog("Internal Error", "A fatal internal error has occured!", e);
                    ed.setVisible(true);
                });
            System.err.println("Could not load missing image icon!");
            System.exit(1);
            return null;
        }
    }
}
