package zander.library.secrets;

import java.io.Serializable;
import java.util.Collections;
import java.util.HashSet;
import java.util.Set;

public abstract class SecretsFactory implements Serializable {
    private static final long serialVersionUID = 3438101849590929517L;

    public abstract boolean isEncrypted();

    public abstract boolean isRecommended();

    public abstract String getBackendName();

    public abstract Secrets createSecrets(String key);

    public abstract boolean isSupported();

    @Override
    public boolean equals(Object o) {
        if (o instanceof SecretsFactory) {
            return ((SecretsFactory) o).getBackendName().equals(getBackendName());
        }
        return false;
    }

    @Override
    public String toString() {
        final StringBuilder sb = new StringBuilder(getBackendName());
        if (isRecommended()) {
            sb.append(" (Recommended)");
        } else if (!isEncrypted()) {
            sb.append(" (Unencrypted, Not Recommended)");
        }
        return sb.toString();
    }

    private static final SecretsFactory DEFAULT_FACTORY = new DefaultSecretsFactory();
    private static Set<SecretsFactory> FACTORIES = Collections.unmodifiableSet(new HashSet<SecretsFactory>());

    public static void registerFactory(SecretsFactory c) {
        final HashSet<SecretsFactory> sf = new HashSet<SecretsFactory>(FACTORIES);
        sf.add(c);
        FACTORIES = Collections.unmodifiableSet(sf);
    }

    public static Set<SecretsFactory> getFactories() {
        return FACTORIES;
    }


    public static SecretsFactory getDefaultFactory() {
        return DEFAULT_FACTORY;
    }

    private static final class DefaultSecretsFactory extends SecretsFactory {
        private static final long serialVersionUID = 8071227730925085343L;

        private SecretsFactory factory;

        public DefaultSecretsFactory() {
            factory = getFactory();
        }

        private SecretsFactory getFactory() {
            if (PassSecretsFactory.isBackendSupported()) {
                return new PassSecretsFactory();
            } else {
                return new PreferencesSecretsFactory();
            }
        }

        @Override
        public boolean isEncrypted() {
            return factory.isEncrypted();
        }

        @Override
        public boolean isRecommended() {
            return factory.isRecommended();
        }

        @Override
        public String getBackendName() {
            return "Default";
        }

        @Override
        public Secrets createSecrets(String key) {
            return factory.createSecrets(key);
        }

        @Override
        public boolean isSupported() {
            return factory != null && factory.isSupported();
        }

        @Override
        public String toString() {
            return "Auto (" + factory.getBackendName() + ")";
        }

    }

}
