package zander.ui;

import java.awt.Dimension;
import java.awt.Window;
import java.awt.image.BufferedImage;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.security.SecureRandom;
import java.util.Objects;

import javax.imageio.ImageIO;
import javax.swing.BoxLayout;
import javax.swing.JComponent;
import javax.swing.JDialog;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JProgressBar;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;
import javax.swing.SwingConstants;
import javax.swing.SwingUtilities;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import zander.library.Library;
import zander.library.LibraryMediaCache;
import zander.library.RemoteStore;
import zander.library.media.ImageMediaData;
import zander.library.media.MediaData;
import zander.library.media.SerializableImageContainer;
import zander.library.media.VideoMediaData;
import zander.library.state.AlbumState;
import zander.library.state.LibraryState;
import zander.library.state.MediaState;

public class LibraryUploadDialog extends JDialog {
    private static final long serialVersionUID = -2859125891849291L;
    private static final Logger LOGGER = LoggerFactory.getLogger(LibraryUploadDialog.class);
    private static final SecureRandom RANDOM = new SecureRandom();
    private static final Dimension SMALL_THUMBNAIL_SIZE = new Dimension(250, 250);
    private static final Dimension MEDIUM_THUMBNAIL_SIZE = new Dimension(500, 500);
    private static final Dimension LARGE_THUMBNAIL_SIZE = new Dimension(1000, 1000);

    private final JLabel headerLabel;
    private final JProgressBar bar;
    private final JTextArea messageArea;
    private final JScrollPane messageScroll;
    private final JPanel content;

    private int progress = 0;

    public LibraryUploadDialog(Window parent) {
        super(parent, "Upload Changes");
        headerLabel = new JLabel("Uploading changes...", SwingConstants.CENTER);
        headerLabel.setAlignmentX(JComponent.CENTER_ALIGNMENT);
        bar = new JProgressBar();
        bar.setAlignmentX(JComponent.CENTER_ALIGNMENT);
        messageArea = new JTextArea();
        messageArea.setEditable(false);
        messageScroll = new JScrollPane(messageArea);
        messageScroll.setAlignmentX(JComponent.CENTER_ALIGNMENT);
        content = new JPanel();
        content.setLayout(new BoxLayout(content, BoxLayout.Y_AXIS));
        content.add(headerLabel);
        content.add(bar);
        content.add(messageScroll);
        setContentPane(content);
        setMinimumSize(new Dimension(UIUtils.SCREEN_SIZE.width / 3, UIUtils.SCREEN_SIZE.height / 3));
        setSize(getMinimumSize());
        setModalityType(JDialog.DEFAULT_MODALITY_TYPE);
        setDefaultCloseOperation(JDialog.DO_NOTHING_ON_CLOSE);
        setResizable(true);
        UIUtils.centerWindow(this, parent);
    }

    private void addMessage(String message) {
        SwingUtilities.invokeLater(() -> {
            String ct = messageArea.getText();
            if (ct.isBlank()) {
                messageArea.setText(message);
            } else {
                messageArea.setText(ct + "\n" + message);
            }
            LOGGER.info(message);
        });
    }

    public boolean upload(LibraryState state, LibraryMediaCache mc) {
        int resp = JOptionPane.showConfirmDialog(getOwner(), "Are you sure you would like to upload changes?" +
                                                 "\nThis process may take a while and cannot be canceled.",
                                                 "Confirm Upload", JOptionPane.YES_NO_OPTION,
                                                 JOptionPane.QUESTION_MESSAGE);
        if (resp != JOptionPane.YES_OPTION) {
            return false;
        }
        bar.setMinimum(0);
        bar.setMaximum(getUploadLength(state));
        Thread ut = new Thread(() -> {
            doUpload(state, mc);
        });
        ut.start();
        setVisible(true);
        return true;
    }

    private void doUpload(LibraryState state, LibraryMediaCache mc) {
        addMessage("Starting upload...");
        Library rl = state.getLibrary();
        RemoteStore rs = rl.getRemoteStore();
        try {
            rs.doFtpActions(() -> {
                addMessage("Moving albums...");
                addMessage("Resolving album movement, 1st pass.");
                for (int i = 0; moveAlbums(state); ++i) {
                    if (i == state.getAlbums().size()) {
                        throw new IOException("Could not resolve album name conflicts");
                    }
                    addMessage("Resolving album movement, " + getNumberSuffixString(i + 2) + " pass.");
                }
                addMessage("Done moving albums.");
                incrementProgress();
                addMessage("Moving media...");
                moveMedia(state);
                addMessage("Done moving media.");
                incrementProgress();
                addMessage("Uploading library config...");
                rl.addAsciiFile(Library.CONFIG_FILE_PATH, LibraryState.toJsonString(state));
                incrementProgress();
                addMessage("Uploading album configs...");
                uploadAlbumConfigs(state);
                addMessage("Uploading media...");
                uploadMedia(state, mc);
                addMessage("Done uploading meadia.");
                addMessage("Deleting old files...");
                deleteOldFiles(state);
                incrementProgress();
                addMessage("Done deleteing old files.");
            });
        } catch (Throwable e) {
            e.printStackTrace();
            SwingUtilities.invokeLater(() -> {
                setVisible(false);
                ErrorDialog ed = new ErrorDialog("Upload Error", "Could not complete upload!", e);
                ed.setVisible(true);
            });
        }
        addMessage("Upload complete. Have a nice day! ^_^");
        SwingUtilities.invokeLater(() -> setVisible(false));
    }

    private void deleteOldFiles(LibraryState state) throws IOException {
        Library rl = state.getLibrary();
        RemoteStore rs = rl.getRemoteStore();
        for (String oan : state.getOriginalAlbumNames()) {
            if (state.getAlbumForName(oan) == null) {
                String ap = getPathForAlbum(oan);
                if (rs.fileExists(ap)) {
                    rl.deleteFile(ap);
                }
            }
        }
        deleteOldMedia(state);
    }

    private void deleteOldMedia(LibraryState state) throws IOException {
        for (AlbumState as : state.getAlbums()) {
            deleteOldMediaForAlbum(state, as);
        }
    }

    private void deleteOldMediaForAlbum(LibraryState state, AlbumState as) throws IOException {
        Library rl = state.getLibrary();
        RemoteStore rs = rl.getRemoteStore();
        for (String omn : as.getOriginalMediaNames()) {
            if (as.getMediaForName(omn) == null) {
                String mp = getPathForMedia(as, omn);
                if (rs.fileExists(mp)) {
                    rl.deleteFile(mp);
                    rs.setIgnoreManifest(true);
                    rs.deleteFile(getSmallThumbnailPathForMedia(as, omn));
                    rs.deleteFile(getLargeThumbnailPathForMedia(as, omn));
                    rs.setIgnoreManifest(false);
                }
            }
        }
    }

    private void moveMedia(LibraryState state) throws IOException {
        for (AlbumState as : state.getAlbums()) {
            addMessage("Resolving media movement for album '" + as.getName() + "', 1st pass.");
            for (int i = 0; moveAlbumMedia(state, as); ++i) {
                if (i == as.getMedia().size()) {
                    throw new IOException("Could not resolve media name conflicts for album '"
                                                   + as.getName() + "'");
                }
                addMessage("Resolving media movement for album '" + as.getName() + "', "
                           + getNumberSuffixString(i + 2)  + " pass.");
            }
            addMessage("Movement resolved for album '" + as.getName() + "'");
        }
    }

    private String getNumberSuffixString(int n) {
        int l = n % 10;
        switch (l) {
        case 1:
            return n + "st";
        case 2:
            return n + "nd";
        case 3:
            return n + "rd";
        default:
            return n + "th";
        }
    }

    private boolean moveAlbumMedia(LibraryState state, AlbumState as) throws IOException {
        Library rl = state.getLibrary();
        RemoteStore rs = rl.getRemoteStore();
        boolean tempFiles = false;
        for (MediaState ms : as.getMedia()) {
            if (ms.getNameOnDisk() == null) {
                continue;
            }
            String nod = ms.getNameOnDisk();
            String nn = ms.getName();
            if (!nod.equals(nn)) {
                if (rs.fileExists(getPathForMedia(as, nn)) && moveMediaToTempFile(state, as, nn)) {
                    tempFiles = true;
                }
                rl.moveFile(getPathForMedia(as, nod), getPathForMedia(as, nn));
                rs.setIgnoreManifest(true);
                rs.moveFile(getSmallThumbnailPathForMedia(as, nod), getSmallThumbnailPathForMedia(as, nn));
                rs.moveFile(getLargeThumbnailPathForMedia(as, nod), getLargeThumbnailPathForMedia(as, nn));
                rs.setIgnoreManifest(false);
                ms.setNameOnDisk(nn);
            }
        }
        return tempFiles;
    }

    private boolean moveMediaToTempFile(LibraryState state, AlbumState as, String nn) throws IOException {
        MediaState ms = null;
        for (MediaState tm : as.getMedia()) {
            if (Objects.equals(nn, tm.getNameOnDisk())) {
                ms = tm;
                break;
            }
        }
        if (ms == null) {
            return false;
        }
        Library rl = state.getLibrary();
        String tn = getTempMediaName(rl, as, nn);
        String tp = getPathForMedia(as, tn);
        String stp = getSmallThumbnailPathForMedia(as, tn);
        String ltp = getLargeThumbnailPathForMedia(as, tn);
        rl.moveFile(getPathForMedia(as, ms.getNameOnDisk()), tp);
        RemoteStore rs = rl.getRemoteStore();
        rs.setIgnoreManifest(true);
        rs.moveFile(getSmallThumbnailPathForMedia(as, ms.getNameOnDisk()), stp);
        rs.moveFile(getLargeThumbnailPathForMedia(as, ms.getNameOnDisk()), ltp);
        rs.setIgnoreManifest(false);
        ms.setNameOnDisk(tn);
        return true;
    }

    private String getSmallThumbnailPathForMedia(AlbumState as, String media) {
        return getSmallThumbnailPathForMedia(as.getNameOnDisk(), media);
    }

    private String getSmallThumbnailPathForMedia(String anod, String media) {
        return getSmallThumbnailPathForAlbum(anod) + media;
    }

    private String getLargeThumbnailPathForMedia(AlbumState as, String media) {
        return getLargeThumbnailPathForMedia(as.getNameOnDisk(), media);
    }

    private String getLargeThumbnailPathForMedia(String anod, String media) {
        return getLargeThumbnailPathForAlbum(anod) + media;
    }

    private String getPathForMedia(AlbumState as, String media) {
        return getPathForMedia(as.getNameOnDisk(), media);
    }

    private String getPathForMedia(String anod, String media) {
        return getPathForAlbum(anod) + media;
    }

    private void uploadMedia(LibraryState state, LibraryMediaCache mc) throws IOException {
        for (AlbumState as : state.getAlbums()) {
            uploadAlbumMedia(state, as, mc);
        }
    }

    private void uploadAlbumMedia(LibraryState state, AlbumState as, LibraryMediaCache mc) throws IOException {
        Library rl = state.getLibrary();
        for (MediaState ms : as.getMedia()) {
            if (ms.getNameOnDisk() == null) {
                MediaData md = mc.getMedia(as.getName(), ms.getName(), null);
                if (md == null) {
                    rl.addBinaryFile(getPathForAlbum(as.getNameOnDisk()) + ms.getName(), new byte[0]);
                } else {
                    byte[] bytes = md.loadRawBytes();
                    rl.addBinaryFile(getPathForAlbum(as.getNameOnDisk()) + ms.getName(), bytes);
                    if (md instanceof ImageMediaData) {
                        uploadImageThumbnail(state, as, mc, ms, bytes);
                    } else if (md instanceof VideoMediaData) {
                        uploadVideoThumbnail(state, as, mc, ms, (VideoMediaData) md);
                    }
                }
                incrementProgress();
            }
        }
    }

    private void uploadVideoThumbnail(LibraryState state, AlbumState as,
                                      LibraryMediaCache mc, MediaState ms, VideoMediaData md) throws IOException {
        SerializableImageContainer bigFrame = md.getThumbnailFrame();
        BufferedImage smallFrame;
        if (as.getThumbnailName().equals(ms.getName())) {
            smallFrame = UIUtils.fitImageToBox(bigFrame.getImage(), MEDIUM_THUMBNAIL_SIZE);
        } else {
            smallFrame = UIUtils.fitImageToBox(bigFrame.getImage(), SMALL_THUMBNAIL_SIZE);
        }
        ByteArrayOutputStream out = new ByteArrayOutputStream();
        ImageIO.write(smallFrame, "png", out);
        Library rl = state.getLibrary();
        RemoteStore rs = rl.getRemoteStore();
        rs.setIgnoreManifest(true);
        rs.addBinaryFile(getSmallThumbnailPathForMedia(as, ms.getName()), out.toByteArray());
        rs.setIgnoreManifest(false);
    }

    private byte[] getScaledImageData(byte[] rawBytes, Dimension d) {
        try {
            ByteArrayInputStream in = new ByteArrayInputStream(rawBytes);
            BufferedImage origImage = ImageIO.read(in);
            BufferedImage scaledImage = UIUtils.fitImageToBox(origImage, d);
            ByteArrayOutputStream out = new ByteArrayOutputStream();
            ImageIO.write(scaledImage, "png", out);
            return out.toByteArray();
        } catch (IOException e) {
            // Should not happen
            throw new RuntimeException(e);
        }
    }

    private void uploadImageThumbnail(LibraryState state, AlbumState as, LibraryMediaCache mc,
                                      MediaState ms, byte[] rawBytes) throws IOException {
        Library rl = state.getLibrary();
        RemoteStore rs = rl.getRemoteStore();
        byte[] smallBytes;
        if (as.getThumbnailName().equals(ms.getName())) {
            smallBytes = getScaledImageData(rawBytes, MEDIUM_THUMBNAIL_SIZE);
        } else {
            smallBytes = getScaledImageData(rawBytes, SMALL_THUMBNAIL_SIZE);
        }
        byte[] largeBytes = getScaledImageData(rawBytes, LARGE_THUMBNAIL_SIZE);
        rs.setIgnoreManifest(true);
        rs.addBinaryFile(getSmallThumbnailPathForMedia(as, ms.getName()), smallBytes);
        rs.addBinaryFile(getLargeThumbnailPathForMedia(as, ms.getName()), largeBytes);
        rs.setIgnoreManifest(false);
    }

    private void uploadAlbumConfigs(LibraryState state) throws IOException {
        Library rl = state.getLibrary();
        for (AlbumState a : state.getAlbums()) {
            rl.addAsciiFile(getPathForAlbum(a.getName()) + Library.ALBUM_CONFIG_FILE_NAME,
                            AlbumState.toJsonString(a));
            incrementProgress();
        }
    }

    private boolean moveAlbums(LibraryState state) throws IOException {
        boolean tempsExist = false;
        Library rl = state.getLibrary();
        for (AlbumState as : state.getAlbums()) {
            RemoteStore rs = rl.getRemoteStore();
            String nod = as.getNameOnDisk();
            String nn = as.getName();
            if (nod == null) {
                as.setNameOnDisk(nn);
            } else if (!nod.equals(nn)) {
                if (rs.fileExists(getPathForAlbum(nn)) && moveAlbumToTempFile(state, nn)) {
                    tempsExist = true;
                }
                rl.moveFile(getPathForAlbum(nod), getPathForAlbum(nn));
                as.setNameOnDisk(nn);
            }
        }
        return tempsExist;
    }

    private boolean moveAlbumToTempFile(LibraryState state, String nod) throws IOException {
        AlbumState as = null;
        for (AlbumState ta : state.getAlbums()) {
            if (Objects.equals(nod, ta.getNameOnDisk())) {
                as = ta;
                break;
            }
        }
        if (as == null) {
            // un untracked file
            return false;
        }
        Library rl = state.getLibrary();
        String tn = getTempAlbumName(rl, as.getName());
        String tp = getPathForAlbum(tn);
        rl.moveFile(getPathForAlbum(as.getNameOnDisk()), tp);
        as.setNameOnDisk(tn);
        return true;
    }

    private String getTempMediaName(Library l, AlbumState as, String cn) throws IOException {
        String tn = getTempName(l, cn + "-", ".tmp");
        RemoteStore rs = l.getRemoteStore();
        if (rs.fileExists(getPathForMedia(as, tn))) {
            try {
                return getTempMediaName(l, as, cn);
            } catch (StackOverflowError e) {
                throw new IOException("Could not generate random file name!");
            }
        }
        return tn;
    }

    private String getTempAlbumName(Library l, String cn) throws IOException {
        String tn = getTempName(l, cn + "-", ".tmp");
        RemoteStore rs = l.getRemoteStore();
        if (rs.fileExists(getPathForAlbum(tn))) {
            try {
                return getTempAlbumName(l, cn);
            } catch (StackOverflowError e) {
                throw new IOException("Could not generate random file name!");
            }
        }
        return tn;
    }

    private String getTempName(Library l, String prefix, String suffix) {
        final StringBuilder sb = new StringBuilder();
        if (prefix != null) {
            sb.append(prefix);
        }
        sb.append(Long.toUnsignedString(RANDOM.nextLong()));
        sb.append(Long.toUnsignedString(RANDOM.nextLong()));
        if (suffix != null) {
            sb.append(suffix);
        }
        return sb.toString();
    }

    private String getPathForAlbum(String name) {
        return Library.ALBUMS_FILE_PATH + "/" + name + "/";
    }

    private String getSmallThumbnailPathForAlbum(String name) {
        return Library.THUMBNAILS_FILE_PATH + "/small/" + name + "/";
    }

    private String getLargeThumbnailPathForAlbum(String name) {
        return Library.THUMBNAILS_FILE_PATH + "/large/" + name + "/";
    }

    private void incrementProgress() {
        SwingUtilities.invokeLater(() -> {
            bar.setValue(++progress);
        });
    }

    private int getUploadLength(LibraryState state) {
        int length = 1; // library config file
        ++length; // move albums
        ++length; // move media
        ++length; // delete files
        length += state.getAlbums().size(); // album config count
        for (AlbumState as : state.getAlbums()) {
            for (MediaState ms : as.getMedia()) {
                if (ms.getNameOnDisk() == null) {
                    ++length; // media that needs to be written to disk
                }
            }
        }
        return length;
    }
}
