package zander.ui.album;

import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.event.FocusAdapter;
import java.awt.event.FocusEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.image.BufferedImage;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.Objects;

import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JTextField;
import javax.swing.SwingConstants;
import javax.swing.SwingUtilities;
import javax.swing.event.DocumentEvent;
import javax.swing.event.DocumentListener;

import zander.library.LibraryMediaCache;
import zander.library.media.ImageMediaComponent;
import zander.library.media.MediaData;
import zander.library.state.AlbumState;
import zander.library.state.LibraryState;
import zander.ui.ErrorDialog;
import zander.ui.UIUtils;
import zander.ui.album.AlbumPropertyUpdateEvent.AlbumEventProperty;

public class AlbumPropertyPanel extends JPanel {
    private static final long serialVersionUID = 2338201447544771877L;

    private final LibraryMediaCache cache;
    private final LibraryState library;
    private final List<AlbumState> openHistory;

    private final JLabel thumbLabel;
    private final JPanel thumbPanel;
    private final JLabel imageCountLabel;
    private final JLabel imageCountNumber;
    private final JLabel thumbNameLabel;
    private final JTextField thumbNameField;
    private final JButton thumbNameReset;
    private final JLabel titleLabel;
    private final JTextField titleField;
    private final JButton titleReset;
    private final JLabel nameLabel;
    private final JTextField nameField;
    private final JButton nameReset;
    private final JCheckBox nameCheck;

    private final ImageMediaComponent missingImageIcon;

    private AlbumState currentAlbum;
    private boolean disableUpdates = false;

    private final ArrayList<AlbumPropertyUpdateListener> propertyListeners = new ArrayList<AlbumPropertyUpdateListener>();

    public AlbumPropertyPanel(LibraryMediaCache cache, List<AlbumState> openHistory) {
        super(new GridBagLayout());
        this.cache = cache;
        this.openHistory = openHistory;
        missingImageIcon = getMissingImageComponent();
        library = cache.getLibrary();
        thumbLabel = new JLabel("Thumbnail:");
        thumbLabel.setHorizontalAlignment(SwingConstants.CENTER);
        thumbPanel = new JPanel(new BorderLayout());
        thumbPanel.setPreferredSize(new Dimension(UIUtils.SCREEN_SIZE.width / 15, UIUtils.SCREEN_SIZE.height / 15));
        thumbPanel.setMinimumSize(new Dimension(UIUtils.SCREEN_SIZE.width / 15, UIUtils.SCREEN_SIZE.height / 15));
        thumbPanel.addMouseListener(new MouseAdapter() {
                @Override
                public void mouseClicked(MouseEvent e) { selectThumbnail(); }
            });
        thumbNameLabel = new JLabel("Thumbnail Name:");
        thumbNameField = new JTextField();
        thumbNameField.getDocument().addDocumentListener(new DocumentListener(){
                @Override
                public void insertUpdate(DocumentEvent e) { doUpdate(); }
                @Override
                public void removeUpdate(DocumentEvent e) { doUpdate(); }
                @Override
                public void changedUpdate(DocumentEvent e) { doUpdate(); }
                private void doUpdate() {
                    if (!disableUpdates) {
                        albumThumbnailUpdated(thumbNameField.getText());
                    }
                }
            });
        thumbNameReset = new JButton("↺");
        thumbNameReset.addActionListener((e) -> {
                thumbNameField.setText(currentAlbum.getOriginalThumbnail());
            });
        imageCountLabel = new JLabel("Images:");
        imageCountNumber = new JLabel("0");
        titleLabel = new JLabel("Album Title:");
        titleField = new JTextField();
        titleField.getDocument().addDocumentListener(new DocumentListener(){
            @Override
            public void insertUpdate(DocumentEvent e) { update(); }
            @Override
            public void removeUpdate(DocumentEvent e) { update(); }
            @Override
            public void changedUpdate(DocumentEvent e) { update(); }

            private void update() {
                titleReset.setEnabled(!resetStringEquals(titleField.getText(), currentAlbum.getOriginalTitle()));
                if (!nameCheck.isSelected()) {
                    disableUpdates = true;
                    nameField.setText(titleField.getText());
                    disableUpdates = false;
                }
            }
        });
        titleField.addActionListener((e) -> {
                if (!disableUpdates) {
                    albumTitleUpdated(titleField.getText());
                }
            });
        titleField.addFocusListener(new FocusAdapter() {
                @Override
                public void focusLost(FocusEvent e) {
                    if (!disableUpdates) {
                        albumTitleUpdated(titleField.getText());
                    }
                }
            });
        titleReset = new JButton("↺");
        titleReset.addActionListener((e) -> {
                titleField.setText(currentAlbum.getOriginalTitle());
                albumTitleUpdated(titleField.getText());
            });
        nameLabel = new JLabel("Custom URL:");
        nameField = new JTextField();
        nameField.getDocument().addDocumentListener(new DocumentListener(){
            @Override
            public void insertUpdate(DocumentEvent e) { update(); }
            @Override
            public void removeUpdate(DocumentEvent e) { update(); }
            @Override
            public void changedUpdate(DocumentEvent e) { update(); }

            private void update() {
                nameReset.setEnabled(!resetStringEquals(nameField.getText(), currentAlbum.getOriginalName()));
            }
        });
        nameField.addActionListener((e) -> {
                if (!disableUpdates) {
                    albumNameUpdated(nameField.getText());
                }
            });
        nameField.addFocusListener(new FocusAdapter() {
                @Override
                public void focusLost(FocusEvent e) {
                    if (!disableUpdates) {
                        albumNameUpdated(nameField.getText());
                    }
                }
            });
        nameReset = new JButton("↺");
        nameReset.addActionListener((e) -> {
                nameField.setText(currentAlbum.getOriginalName());
                albumNameUpdated(nameField.getText());
            });
        nameCheck = new JCheckBox("Different Title and URL");
        nameCheck.addActionListener((e) -> {
                if (!disableUpdates) {
                    boolean en = nameCheck.isSelected();
                    if (!en) {
                        String name = titleField.getText();
                        if (checkNameErrorAndShowMessage(name)) {
                            disableUpdates = true;
                            nameCheck.setSelected(true);
                            disableUpdates = false;
                            return;
                        } else {
                            String oldName = currentAlbum.getName();
                            currentAlbum.setName(name);
                            cache.moveAlbum(oldName, name);
                            disableUpdates = true;
                            nameField.setText(name);
                            disableUpdates = false;
                        }
                    }
                    currentAlbum.setNameFollowsTitle(!en);
                    nameReset.setEnabled(en);
                    nameField.setEnabled(en);
                    nameLabel.setEnabled(en);
                }
            });
        buildUI();
    }

    private ImageMediaComponent getMissingImageComponent() {
        try {
            BufferedImage bi = UIUtils.loadBufferedImage(ClassLoader.getSystemResource("missing-image-icon.svg"), -1, -1);
            return new ImageMediaComponent(bi);
        } catch (IOException e) {
            ErrorDialog ed = new ErrorDialog("Internal Error", "A fatal error has occured!", e);
            ed.setVisible(true);
            System.exit(0);
            return null;
        }
    }

    private void buildUI() {
        GridBagConstraints c = new GridBagConstraints();
        c.insets = new Insets(1, 1, 1, 1);
        c.gridwidth = 1;
        c.gridheight = 1;
        c.gridx = 0;
        c.gridy = 0;
        c.weighty = 0.1;
        c.weightx = 0.0;
        c.anchor = GridBagConstraints.LINE_END;
        c.fill = GridBagConstraints.NONE;
        add(thumbLabel, c);
        c.gridx = 1;
        c.anchor = GridBagConstraints.LINE_START;
        c.fill = GridBagConstraints.BOTH;
        c.weightx = 1.0;
        c.weighty = 1.0;
        add(thumbPanel, c);
        c.weighty = 0.1;
        c.gridx = 0;
        c.gridy = 1;
        c.anchor = GridBagConstraints.LINE_END;
        c.fill = GridBagConstraints.NONE;
        c.weightx = 0.0;
        add(imageCountLabel, c);
        c.gridx = 1;
        c.anchor = GridBagConstraints.LINE_START;
        add(imageCountNumber, c);
        c.gridwidth = 1;
        c.gridx = 0;
        c.gridy = 2;
        c.anchor = GridBagConstraints.LINE_END;
        add(thumbNameLabel, c);
        c.gridx = 1;
        c.anchor = GridBagConstraints.CENTER;
        c.fill = GridBagConstraints.HORIZONTAL;
        add(thumbNameField, c);
        c.fill = GridBagConstraints.NONE;
        c.anchor = GridBagConstraints.LINE_START;
        c.gridx = 2;
        add(thumbNameReset, c);
        c.gridwidth = 1;
        c.gridx = 0;
        c.gridy = 3;
        c.anchor = GridBagConstraints.LINE_END;
        add(titleLabel, c);
        c.gridx = 1;
        c.anchor = GridBagConstraints.CENTER;
        c.fill = GridBagConstraints.HORIZONTAL;
        add(titleField, c);
        c.fill = GridBagConstraints.NONE;
        c.anchor = GridBagConstraints.LINE_START;
        c.gridx = 2;
        add(titleReset, c);
        c.gridx = 0;
        c.gridy = 4;
        c.anchor = GridBagConstraints.LINE_END;
        add(nameLabel, c);
        c.gridx = 1;
        c.anchor = GridBagConstraints.CENTER;
        c.fill = GridBagConstraints.HORIZONTAL;
        add(nameField, c);
        c.anchor = GridBagConstraints.LINE_START;
        c.fill = GridBagConstraints.NONE;
        c.gridx = 2;
        add(nameReset, c);
        c.gridx = 0;
        c.gridy = 5;
        c.gridwidth = 3;
        c.anchor = GridBagConstraints.CENTER;
        add(nameCheck, c);
    }

    public void setAlbum(AlbumState album) {
        setAlbum(album, false);
    }

    public void setAlbum(AlbumState album, boolean clearOld) {
        disableUpdates = true;
        if (currentAlbum != null && !clearOld) {
            if (!titleField.getText().equals(currentAlbum.getTitle())) {
                currentAlbum.setTitle(titleField.getText());
            }
            if (!nameField.getText().equals(currentAlbum.getName())) {
                String name = nameField.getText();
                if (!checkNameErrorAndShowMessage(name)) {
                    String oldName = currentAlbum.getName();
                    currentAlbum.setName(name);
                    cache.moveAlbum(oldName, name);
                } else {
                    currentAlbum.setTitle(currentAlbum.getName());
                }
            }
        }
        currentAlbum = album;
        refreshThumbnail();
        thumbNameField.setText(album.getThumbnailName());
        titleField.setText(album.getTitle());
        nameField.setText(album.getName());
        boolean nft = album.doesNameFollowTitle();
        nameCheck.setSelected(!nft);
        nameField.setEnabled(!nft);
        nameLabel.setEnabled(!nft);
        nameReset.setEnabled(!nft);
        imageCountNumber.setText(Integer.toString(album.getMedia().size()));
        updateResetButtons();
        disableUpdates = false;
    }

    public AlbumState getAlbum() {
        return currentAlbum;
    }

    public String getThumbnailName() {
        return thumbNameField.getText();
    }

    public void setThumbnailName(String name) {
        thumbNameField.setText(name);
    }

    public String getAlbumName() {
        return nameField.getText();
    }

    public void setAlbumName(String name) {
        nameField.setText(name);
    }

    public String getAlbumTitle() {
        return titleField.getText();
    }

    public void setAlbumTitle(String title) {
        titleField.setText(title);
    }

    public LibraryMediaCache getMediaCache() {
        return cache;
    }

    public LibraryState getLibrary() {
        return library;
    }

    public void addAlbumPropertyUpdateListener(AlbumPropertyUpdateListener l) {
        propertyListeners.add(l);
    }

    public void removeAlbumPropertyUpdateListener(AlbumPropertyUpdateListener l) {
        propertyListeners.remove(l);
    }

    private void fireUpdatePropertyEvent(String oldVal, String newVal, AlbumEventProperty prop) {
        AlbumPropertyUpdateEvent e = new AlbumPropertyUpdateEvent(currentAlbum, oldVal, newVal, prop);
        for (AlbumPropertyUpdateListener l : propertyListeners) {
            l.propertyUpdated(e);
        }
    }

    private void selectThumbnail() {
        AlbumThumbnailSelectDialog diag =
            new AlbumThumbnailSelectDialog(SwingUtilities.getWindowAncestor(this), cache, currentAlbum, openHistory);
        diag.setVisible(true);
        if (diag.getResponse() == AlbumThumbnailSelectDialog.RESPONSE_SELECT && diag.getSelectedThumbnail() != null) {
            thumbNameField.setText(diag.getSelectedThumbnail());
        }
    }

    private void albumThumbnailUpdated(String thumbnail) {
        String oldName = currentAlbum.getThumbnailName();
        currentAlbum.setThumbnailName(thumbnail);
        refreshThumbnail();
        thumbNameReset.setEnabled(!resetStringEquals(thumbnail, currentAlbum.getOriginalThumbnail()));
        fireUpdatePropertyEvent(oldName, thumbnail, AlbumEventProperty.THUMBNAIL);
    }

    private void albumTitleUpdated(String title) {
        if (!nameCheck.isSelected()) {
            if (!title.equals(currentAlbum.getName()) && library.hasAlbum(title)) {
                disableUpdates = true;
                titleField.setText(currentAlbum.getName());
                disableUpdates = false;
                JOptionPane.showMessageDialog(this, "An album with the title \"" + title + "\" is invalid becuase it " +
                                              "result in a URL that is the same as one that exists!",
                                              "Name Conflict", JOptionPane.ERROR_MESSAGE);
                return;
            } else if (title.isBlank()) {
                disableUpdates = true;
                titleField.setText(currentAlbum.getName());
                disableUpdates = false;
                JOptionPane.showMessageDialog(this, "The title \"" + title + "\" is invalid becaue it would result in an " +
                                              " invalid URL!",
                                              "Invalid Name", JOptionPane.ERROR_MESSAGE);
                return;
            } else if (title.contains("/")) {
                disableUpdates = true;
                titleField.setText(currentAlbum.getName());
                disableUpdates = false;
                JOptionPane.showMessageDialog(this, "The title \"" + title +
                                              "\" is invalid because it contains the character '/' which is not " +
                                              "allowed in URLs!",
                                              "Invalid Name", JOptionPane.ERROR_MESSAGE);
                return;
            }
            String oldName = currentAlbum.getName();
            currentAlbum.setName(title);
            cache.moveAlbum(oldName, title);
        }
        String oldTitle = currentAlbum.getTitle();
        currentAlbum.setTitle(title);
        titleReset.setEnabled(!resetStringEquals(title, currentAlbum.getOriginalTitle()));
        fireUpdatePropertyEvent(oldTitle, title, AlbumEventProperty.TITLE);
    }

    /** @return true if an error occured, false otherwise */
    private boolean checkNameErrorAndShowMessage(String name) {
        if (name.equals(currentAlbum.getName())) {
            return false;
        } else if (library.hasAlbum(name)) {
            disableUpdates = true;
            nameField.setText(currentAlbum.getName());
            disableUpdates = false;
            JOptionPane.showMessageDialog(this, "An album with the URL \"" + name + "\" already exists!",
                                            "Name Conflict", JOptionPane.ERROR_MESSAGE);
            return true;
        } else if (name.isBlank()) {
            disableUpdates = true;
            nameField.setText(currentAlbum.getName());
            disableUpdates = false;
            JOptionPane.showMessageDialog(this, "The URL \"" + name + "\" is invalid!",
                                            "Invalid Name", JOptionPane.ERROR_MESSAGE);
            return true;
        } else if (name.contains("/")) {
            disableUpdates = true;
            nameField.setText(currentAlbum.getName());
            disableUpdates = false;
            JOptionPane.showMessageDialog(this,
                    "The URL \"" + name + "\" is invalid because it contains the character '/'!", "Invalid Name",
                    JOptionPane.ERROR_MESSAGE);
            return true;
        }
        return false;
    }

    private void albumNameUpdated(String name) {
        if (!currentAlbum.getName().equals(name)) {
            if (!checkNameErrorAndShowMessage(name)) {
                String oldName = currentAlbum.getName();
                currentAlbum.setName(name);
                cache.moveAlbum(oldName, name);
            }
        }
        nameReset.setEnabled(!resetStringEquals(nameField.getText(), currentAlbum.getOriginalName()));
    }

    private void refreshThumbnail() {
        MediaData md = cache.getCheckedMedia(currentAlbum.getName(), currentAlbum.getThumbnailName(), null);
        thumbPanel.removeAll();
        if (md != null) {
            Component thumbComp = md.createThumbnailComponent();
            thumbPanel.add(thumbComp, BorderLayout.CENTER);
        } else {
            thumbPanel.add(missingImageIcon, BorderLayout.CENTER);
        }
    }

    private void updateResetButtons() {
        thumbNameReset.setEnabled(!resetStringEquals(thumbNameField.getText(), currentAlbum.getOriginalThumbnail()));
        nameReset.setEnabled(!resetStringEquals(nameField.getText(), currentAlbum.getOriginalName()) &&
            nameCheck.isSelected());
        titleReset.setEnabled(!resetStringEquals(titleField.getText(), currentAlbum.getOriginalTitle()));
    }

    private static boolean resetStringEquals(String s1, String s2) {
        if (s1 != null && s1.isEmpty()) {
            s1 = null;
        }
        if (s2 != null && s2.isEmpty()) {
            s2 = null;
        }
        return Objects.equals(s1, s2);
    }

    public void updateName() {
        albumNameUpdated(nameField.getText());
    }
}
