package zander.ui.cellrenderer;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.image.BufferedImage;
import java.io.IOException;

import javax.swing.BorderFactory;
import javax.swing.BoxLayout;
import javax.swing.JLayeredPane;
import javax.swing.JList;
import javax.swing.JPanel;
import javax.swing.JTextPane;
import javax.swing.ListCellRenderer;
import javax.swing.text.SimpleAttributeSet;
import javax.swing.text.StyleConstants;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import zander.library.LibraryMediaCache;
import zander.library.media.ImageMediaComponent;
import zander.library.media.MediaData;
import zander.library.media.SerializableImageContainer;
import zander.library.state.AlbumState;
import zander.library.state.LibraryState;
import zander.library.state.MediaState;
import zander.ui.ErrorDialog;
import zander.ui.OverlayPainterComponent;
import zander.ui.UIUtils;

public class MediaCellRenderer extends JPanel implements ListCellRenderer<MediaState> {
    private static final long serialVersionUID = 8224231026387257322L;
    private static final Logger LOGGER = LoggerFactory.getLogger(MediaCellRenderer.class);
    private static final SerializableImageContainer STAR_IMAGE = UIUtils
            .loadSerializableImage(ClassLoader.getSystemResource("star.svg"), -1, -1);

    private final LibraryMediaCache cache;
    private final AlbumState album;
    private final LibraryState library;
    private final JLayeredPane layeredPane;
    private final JPanel compPanel;
    private final OverlayPainterComponent overlay;
    private final ImageMediaComponent missingImageIcon;
    private final JTextPane name;
    private final SimpleAttributeSet attribs;

    public MediaCellRenderer(LibraryMediaCache cache, AlbumState album) {
        setLayout(new BorderLayout());
        setBorder(BorderFactory.createEmptyBorder(4, 4, 4, 4));
        this.cache = cache;
        this.album = album;
        this.library = cache.getLibrary();
        layeredPane = new JLayeredPane();
        add(layeredPane, BorderLayout.CENTER);
        compPanel = new JPanel();
        compPanel.setLayout(new BoxLayout(compPanel, BoxLayout.Y_AXIS));
        overlay = new OverlayPainterComponent(STAR_IMAGE, 0, 0, 0, 0);
        missingImageIcon = getMissingImageComponent();
        name = new JTextPane();
        name.setOpaque(false);
        attribs = new SimpleAttributeSet();
        StyleConstants.setAlignment(attribs, StyleConstants.ALIGN_CENTER);
        layeredPane.add(compPanel, Integer.valueOf(0));
        layeredPane.add(overlay, Integer.valueOf(1));
    }

    private ImageMediaComponent getMissingImageComponent() {
        try {
            BufferedImage bi = UIUtils.loadBufferedImage(ClassLoader.getSystemResource("missing-image-icon.svg"), -1, -1);
            return new ImageMediaComponent(bi);
        } catch (IOException e) {
            LOGGER.error("Could not load missing image icon!", e);
            ErrorDialog ed = new ErrorDialog("Internal Error", "A fatal error has occured!", e);
            ed.setVisible(true);
            System.exit(0);
            return null;
        }
    }

    public LibraryState getLibrary() {
        return library;
    }

    @Override
    public void setFont(Font font) {
        super.setFont(font);
        if (name != null) {
            name.setFont(font);
        }
    }

    @Override
    public void setForeground(Color fg) {
        super.setForeground(fg);
        if (name != null) {
            name.setForeground(fg);
        }
    }

    @Override
    public void setBackground(Color bg) {
        super.setBackground(bg);
        if (name != null) {
            name.setBackground(bg);
        }
    }

    private void setNameLabel(MediaState state) {
        String text = state.getName();
        this.name.setText(text);
        this.name.getStyledDocument().setParagraphAttributes(0, text.length(), attribs, false);
    }

    @Override
    public Component getListCellRendererComponent(JList<? extends MediaState> list, MediaState value, int index,
                                                  boolean isSelected, boolean cellHasFocus) {
        Color bg;
        Color fg;
        if (isSelected) {
            bg = list.getSelectionBackground();
            fg = list.getSelectionForeground();
        } else {
            bg = list.getBackground();
            fg = list.getForeground();
        }
        setBackground(bg);
        setForeground(fg);
        compPanel.setBackground(bg);
        compPanel.setForeground(fg);
        setFont(list.getFont());
        setNameLabel(value);
        MediaData m = cache.getMedia(album.getName(), value.getName(), null);
        Component c;
        if (m != null) {
            c = m.createThumbnailComponent();
        } else {
            c = missingImageIcon;
        }
        c.setPreferredSize(UIUtils.THUMBNAIL_SIZE);
        name.setPreferredSize(new Dimension(Math.round(c.getPreferredSize().width * 1.5f), c.getPreferredSize().height / 3));
        compPanel.removeAll();
        compPanel.add(c);
        compPanel.add(name);
        Dimension ns = compPanel.getLayout().preferredLayoutSize(compPanel);
        setSize(ns);
        setPreferredSize(ns);
        compPanel.setLocation(0, 0);
        compPanel.setSize(getPreferredSize());
        if (album.getThumbnailName().equals(value.getName())) {
            overlay.setBounds(compPanel.getBounds());
            int scd = Math.min(getWidth(), getHeight());
            int op = (int) Math.floor(scd * 0.05f);
            int os = (int) Math.floor(scd * 0.15f);
            overlay.setOverlayBounds(op, op, os, os);
            overlay.setVisible(true);
        } else {
            overlay.setVisible(false);
        }
        return this;
    }

}
