package zander.ui.media;

import java.awt.Window;
import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.FocusAdapter;
import java.awt.event.FocusEvent;
import java.awt.image.BufferedImage;
import java.io.IOException;
import java.text.NumberFormat;
import java.time.LocalDate;
import java.time.LocalTime;
import java.util.Objects;

import javax.swing.JButton;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;
import javax.swing.JTextField;
import javax.swing.SwingUtilities;
import javax.swing.UIManager;
import javax.swing.event.DocumentEvent;
import javax.swing.event.DocumentListener;

import com.github.lgooddatepicker.components.DatePicker;
import com.github.lgooddatepicker.components.DatePickerSettings;
import com.github.lgooddatepicker.components.DatePickerSettings.DateArea;
import com.github.lgooddatepicker.components.TimePicker;
import com.github.lgooddatepicker.components.TimePickerSettings;
import com.github.lgooddatepicker.components.TimePickerSettings.TimeArea;

import org.jxmapviewer.viewer.GeoPosition;

import zander.library.LibraryMediaCache;
import zander.library.media.ImageMediaComponent;
import zander.library.media.MediaData;
import zander.library.state.AlbumState;
import zander.library.state.MediaState;
import zander.ui.ErrorDialog;
import zander.ui.UIUtils;
import zander.ui.media.map.LocationSelectDialog;

public class MediaPropertyPanel extends JPanel {
    private static final long serialVersionUID = -360830663146686639L;

    private static final NumberFormat NUM_FORMAT = NumberFormat.getInstance();
    private static final long GB_SIZE = (long) Math.pow(1024, 3);
    private static final long MB_SIZE = (long) Math.pow(1024, 2);
    private static final long KB_SIZE = 1024;

    static {
        NUM_FORMAT.setGroupingUsed(false);
        NUM_FORMAT.setMinimumFractionDigits(0);
        NUM_FORMAT.setMaximumFractionDigits(2);
    }

    private final LibraryMediaCache cache;
    private final AlbumState album;

    private final JLabel previewLabel;
    private final JPanel previewPanel;
    private final JLabel thumbLabel;
    private final JLabel thumbValue;
    private final JLabel sizeLabel;
    private final JLabel sizeValue;
    private final JLabel typeLabel;
    private final JLabel typeValue;
    private final JLabel nameLabel;
    private final JTextField nameField;
    private final JButton nameReset;
    private final JLabel descLabel;
    private final JTextArea descArea;
    private final JScrollPane descScroll;
    private final JPanel descPanel;
    private final JButton descReset;
    private final JLabel dateLabel;
    private final DatePicker datePicker;
    private final JButton dateReset;
    private final JLabel timeLabel;
    private final TimePicker timePicker;
    private final JButton timeReset;
    private final JLabel mapLabel;
    private final JLabel mapValue;
    private final JButton mapButton;

    private final ImageMediaComponent missingImageIcon;

    private MediaState currentMedia;
    private boolean disableUpdates = false;

    public MediaPropertyPanel(LibraryMediaCache cache, AlbumState album) {
        super(new GridBagLayout());
        this.cache = cache;
        this.album = album;
        missingImageIcon = getMissingImageComponent();
        previewLabel = new JLabel("Preview:");
        previewPanel = new JPanel(new BorderLayout());
        thumbLabel = new JLabel("Is Thumbnail:");
        thumbValue = new JLabel("No");
        sizeLabel = new JLabel("File Size:");
        sizeValue = new JLabel("N/A");
        typeLabel = new JLabel("File Type:");
        typeValue = new JLabel("unknown");
        nameLabel = new JLabel("Name:");
        nameField = new JTextField();
        nameField.getDocument().addDocumentListener(new DocumentListener(){
            @Override
            public void insertUpdate(DocumentEvent e) { update(); }
            @Override
            public void removeUpdate(DocumentEvent e) { update(); }
            @Override
            public void changedUpdate(DocumentEvent e) { update(); }

            private void update() {
                nameReset.setEnabled(!Objects.equals(nameField.getText(), currentMedia.getOriginalName()));
            }
        });
        nameField.addActionListener((e) -> {
                if (!disableUpdates) {
                    nameChanged(nameField.getText());
                }
            });
        nameField.addFocusListener(new FocusAdapter() {
                @Override
                public void focusLost(FocusEvent e) {
                    if (!disableUpdates) {
                        nameChanged(nameField.getText());
                    }
                }
            });
        nameReset = new JButton("↺");
        nameReset.addActionListener((e) -> {
                nameField.setText(currentMedia.getOriginalName());
                nameChanged(nameField.getText());
            });
        descLabel = new JLabel("Description:");
        descArea = new JTextArea();
        descArea.setLineWrap(true);
        descArea.setBorder(null);
        descArea.getDocument().addDocumentListener(new DocumentListener() {
                @Override
                public void insertUpdate(DocumentEvent e) { update(); }
                @Override
                public void removeUpdate(DocumentEvent e) { update(); }
                @Override
                public void changedUpdate(DocumentEvent e) { update(); }

                private void update() {
                    if (!disableUpdates) {
                        descriptionChanged(descArea.getText());
                    }
                }
            });
        descScroll = new JScrollPane(descArea);
        descScroll.setBorder(null);
        descPanel = new JPanel(new BorderLayout());
        descPanel.setBorder(UIManager.getBorder("TextField.border"));
        descPanel.add(descScroll, BorderLayout.CENTER);
        descReset = new JButton("↺");
        descReset.addActionListener((e) -> {
                descArea.setText(currentMedia.getOriginalDescription());
            });
        dateLabel = new JLabel("Date:");
        datePicker = new DatePicker(getDatePickerSettings());
        datePicker.addDateChangeListener((e) -> {
                if (!disableUpdates) {
                    dateChanged(datePicker.getDate());
                }
            });
        dateReset = new JButton("↺");
        dateReset.addActionListener((e) -> {
                datePicker.setDate(currentMedia.getOriginalDate());
            });
        timeLabel = new JLabel("Time:");
        timePicker = new TimePicker(getTimePickerSettings());
        timePicker.addTimeChangeListener((e) -> {
                if (!disableUpdates) {
                    timeChanged(timePicker.getTime());
                }
            });
        timeReset = new JButton("↺");
        timeReset.addActionListener((e) -> {
                timePicker.setTime(currentMedia.getOriginalTime());
            });
        mapLabel = new JLabel("Location:");
        mapValue = new JLabel("N/A");
        mapButton = new JButton("\u2026");
        mapButton.addActionListener(this::mapButtonHandler);
        buildUI();
    }

    private Color getDatePickerWeekHeaderColor() {
        Color sel = UIManager.getColor("List.selectionBackground");
        if (sel == null) {
            return Color.white;
        }
        return UIUtils.scaleColor(sel, 0.9f, 0.9f, 0.9f);
    }

    private DatePickerSettings getDatePickerSettings() {
        DatePickerSettings s = new DatePickerSettings();
        s.setAllowEmptyDates(true);
        s.setAllowKeyboardEditing(true);
        s.setColorBackgroundWeekdayLabels(getDatePickerWeekHeaderColor(), true);
        s.setColor(DateArea.BackgroundCalendarPanelLabelsOnHover, UIManager.getColor("List.selectionBackground"));
        s.setColor(DateArea.BackgroundClearLabel, UIManager.getColor("Panel.background"));
        s.setColor(DateArea.BackgroundMonthAndYearMenuLabels, UIManager.getColor("Panel.background"));
        s.setColor(DateArea.BackgroundMonthAndYearNavigationButtons, UIManager.getColor("Panel.background"));
        s.setColor(DateArea.BackgroundOverallCalendarPanel, UIManager.getColor("Panel.background"));
        s.setColor(DateArea.BackgroundTodayLabel, UIManager.getColor("Panel.background"));
        s.setColor(DateArea.BackgroundTopLeftLabelAboveWeekNumbers, UIManager.getColor("Panel.background"));
        s.setColor(DateArea.CalendarBackgroundNormalDates, UIManager.getColor("Panel.background"));
        s.setColor(DateArea.CalendarBackgroundSelectedDate, UIManager.getColor("List.selectionBackground"));
        s.setColor(DateArea.CalendarBorderSelectedDate, UIManager.getColor("Tree.selectionBorderColor"));
        s.setColor(DateArea.CalendarDefaultBackgroundHighlightedDates, UIManager.getColor("Panel.background"));
        s.setColor(DateArea.CalendarDefaultTextHighlightedDates, UIManager.getColor("List.selectionBackground"));
        s.setColor(DateArea.CalendarTextNormalDates, UIManager.getColor("Label.foreground"));
        s.setColor(DateArea.CalendarTextWeekNumbers, UIManager.getColor("Label.foreground"));
        s.setColor(DateArea.CalendarTextWeekdays, UIManager.getColor("Label.foreground"));
        s.setColor(DateArea.DatePickerTextValidDate, UIManager.getColor("Label.foreground"));
        s.setColor(DateArea.TextCalendarPanelLabelsOnHover, UIManager.getColor("List.selectionForeground"));
        s.setColor(DateArea.TextFieldBackgroundValidDate, UIManager.getColor("TextField.background"));
        s.setColor(DateArea.TextMonthAndYearMenuLabels, UIManager.getColor("Label.foreground"));
        s.setColor(DateArea.TextMonthAndYearNavigationButtons, UIManager.getColor("Label.foreground"));
        s.setColor(DateArea.TextTodayLabel, UIManager.getColor("Label.foreground"));
        return s;
    }

    private TimePickerSettings getTimePickerSettings() {
        TimePickerSettings s = new TimePickerSettings();
        s.setAllowEmptyTimes(true);
        s.setAllowKeyboardEditing(true);
        s.setColor(TimeArea.TextFieldBackgroundValidTime, UIManager.getColor("TextField.background"));
        s.setColor(TimeArea.TimePickerTextValidTime, UIManager.getColor("Label.foreground"));
        return s;
    }

    private void buildUI() {
        GridBagConstraints gbc = new GridBagConstraints();
        gbc.insets = new Insets(1, 1, 1, 1);
        gbc.weightx = 0.0;
        gbc.weighty = 0.1;
        gbc.gridwidth = 1;
        gbc.gridheight = 1;
        gbc.gridx = 0;
        gbc.gridy = 0;
        gbc.anchor = GridBagConstraints.LINE_END;
        gbc.fill = GridBagConstraints.NONE;
        add(previewLabel, gbc);
        gbc.gridx = 1;
        gbc.weightx = 1.0;
        gbc.weighty = 1.0;
        gbc.fill = GridBagConstraints.BOTH;
        gbc.anchor = GridBagConstraints.CENTER;
        add(previewPanel, gbc);
        gbc.weightx = 0.0;
        gbc.weighty = 0.1;
        gbc.fill = GridBagConstraints.NONE;
        gbc.anchor = GridBagConstraints.LINE_END;
        gbc.gridx = 0;
        gbc.gridy = 1;
        add(thumbLabel, gbc);
        gbc.gridx = 1;
        gbc.anchor = GridBagConstraints.LINE_START;
        add(thumbValue, gbc);
        gbc.anchor = GridBagConstraints.LINE_END;
        gbc.gridx = 0;
        gbc.gridy = 2;
        add(sizeLabel, gbc);
        gbc.gridx = 1;
        gbc.anchor = GridBagConstraints.LINE_START;
        add(sizeValue, gbc);
        gbc.anchor = GridBagConstraints.LINE_END;
        gbc.gridx = 0;
        gbc.gridy = 3;
        add(typeLabel, gbc);
        gbc.gridx = 1;
        gbc.anchor = GridBagConstraints.LINE_START;
        add(typeValue, gbc);
        gbc.gridx = 0;
        gbc.gridy = 4;
        gbc.anchor = GridBagConstraints.LINE_END;
        add(nameLabel, gbc);
        gbc.fill = GridBagConstraints.HORIZONTAL;
        gbc.gridx = 1;
        gbc.anchor = GridBagConstraints.CENTER;
        add(nameField, gbc);
        gbc.fill = GridBagConstraints.NONE;
        gbc.gridx = 2;
        gbc.anchor = GridBagConstraints.LINE_START;
        add(nameReset, gbc);
        gbc.gridx = 0;
        gbc.gridy = 5;
        gbc.anchor = GridBagConstraints.LINE_END;
        add(descLabel, gbc);
        gbc.gridx = 1;
        gbc.fill = GridBagConstraints.BOTH;
        gbc.anchor = GridBagConstraints.CENTER;
        gbc.weighty = 0.3;
        add(descPanel, gbc);
        gbc.weighty = 0.1;
        gbc.fill = GridBagConstraints.NONE;
        gbc.gridx = 2;
        gbc.anchor = GridBagConstraints.LINE_START;
        gbc.weightx = 0.0;
        add(descReset, gbc);
        gbc.gridx = 0;
        gbc.gridy = 6;
        gbc.anchor = GridBagConstraints.LINE_END;
        add(dateLabel, gbc);
        gbc.gridx = 1;
        gbc.fill = GridBagConstraints.HORIZONTAL;
        gbc.anchor = GridBagConstraints.CENTER;
        add(datePicker, gbc);
        gbc.fill = GridBagConstraints.NONE;
        gbc.gridx = 2;
        gbc.anchor = GridBagConstraints.LINE_START;
        add(dateReset, gbc);
        gbc.gridx = 0;
        gbc.gridy = 7;
        gbc.anchor = GridBagConstraints.LINE_END;
        add(timeLabel, gbc);
        gbc.gridx = 1;
        gbc.fill = GridBagConstraints.HORIZONTAL;
        gbc.anchor = GridBagConstraints.CENTER;
        add(timePicker, gbc);
        gbc.fill = GridBagConstraints.NONE;
        gbc.gridx = 2;
        gbc.anchor = GridBagConstraints.LINE_START;
        add(timeReset, gbc);
        gbc.fill = GridBagConstraints.NONE;
        gbc.anchor = GridBagConstraints.LINE_END;
        gbc.gridx = 0;
        gbc.gridy = 8;
        add(mapLabel, gbc);
        gbc.gridx = 1;
        gbc.fill = GridBagConstraints.HORIZONTAL;
        gbc.anchor = GridBagConstraints.CENTER;
        add(mapValue, gbc);
        gbc.gridx = 2;
        gbc.fill = GridBagConstraints.HORIZONTAL;
        gbc.anchor = GridBagConstraints.CENTER;
        add(mapButton, gbc);
    }

    public LibraryMediaCache getMediaCache() {
        return cache;
    }

    public MediaState getCurrentMedia() {
        return currentMedia;
    }

    private static String getFileSizeString(long size) {
        double ds = (double) size;
        if (size >= GB_SIZE) {
            return NUM_FORMAT.format(ds / GB_SIZE) + " GB";
        } else if (size >= MB_SIZE) {
            return NUM_FORMAT.format(ds / MB_SIZE) + " MB";
        } else if (size >= KB_SIZE) {
            return NUM_FORMAT.format(ds / KB_SIZE) + " KB";
        } else {
            return size + " B";
        }
    }

    public void setCurrentMedia(MediaState media) {
        disableUpdates = true;
        if (currentMedia != null && !nameField.getText().equals(currentMedia.getName())) {
            nameChanged(nameField.getText());
        }
        this.currentMedia = media;
        updateIsThumbnail();
        sizeValue.setText(getFileSizeString(media.getFileSize()));
        nameField.setText(media.getName());
        descArea.setText(media.getDescription());
        previewPanel.removeAll();
        MediaData md = cache.getMedia(album.getName(), media.getName(), null);
        if (md != null) {
            typeValue.setText(md.getType());
            md.getViewerComponent().setEnabled(true);
            previewPanel.add(md.getViewerComponent(), BorderLayout.CENTER);
        } else {
            typeValue.setText("unknown");
            previewPanel.add(missingImageIcon, BorderLayout.CENTER);
        }
        datePicker.setDate(media.getDate());
        timePicker.setTime(media.getTime());
        mapValue.setText(formatLocation(media));
        revalidate();
        repaint();
        updateResetButtons();
        disableUpdates = false;
    }

    public void updateIsThumbnail() {
        if (currentMedia != null) {
            if (album.getThumbnailName().equals(currentMedia.getName())) {
                thumbValue.setText("Yes");
            } else {
                thumbValue.setText("No");
            }
        }
    }

    private static String formatLocation(MediaState m) {
        if (m.getLatitude() == null || m.getLongitude() == null) {
            return "N/A";
        }
        return m.getLatitude() + ", " + m.getLongitude();
    }

    private void nameChanged(String name) {
        String oldName = currentMedia.getName();
        if (!name.equals(oldName)) {
            if (album.getMediaForName(name) != null) {
                disableUpdates = true;
                nameField.setText(oldName);
                disableUpdates = false;
                JOptionPane.showMessageDialog(this, "Media with the name \"" + name + "\" already exists!",
                        "Name Conflict", JOptionPane.ERROR_MESSAGE);
            } else if (name.isBlank()) {
                disableUpdates = true;
                nameField.setText(oldName);
                disableUpdates = false;
                JOptionPane.showMessageDialog(this, "The name \"" + name + "\" is invalid!", "Invalid Name",
                        JOptionPane.ERROR_MESSAGE);
            }  else if (name.equals("data.json")) {
                disableUpdates = true;
                nameField.setText(oldName);
                disableUpdates = false;
                JOptionPane.showMessageDialog(this, "The name \"" + name + "\" is reserved!", "Invalid Name",
                        JOptionPane.ERROR_MESSAGE);
            } else if (name.contains("/")) {
                disableUpdates = true;
                nameField.setText(oldName);
                disableUpdates = false;
                JOptionPane.showMessageDialog(this, "The name \"" + name + "\" is because the charater '/' is not allowed!",
                                              "Invalid Name", JOptionPane.ERROR_MESSAGE);
            } else {
                currentMedia.setName(name);
                cache.moveMedia(album.getName(), oldName, name);
                if (album.getThumbnailName().equals(oldName)) {
                    album.setThumbnailName(name);
                }
                SwingUtilities.getWindowAncestor(this).repaint();
            }
        }
        nameReset.setEnabled(!Objects.equals(nameField.getText(), currentMedia.getOriginalName()));
    }

    private void descriptionChanged(String description) {
        currentMedia.setDescription(description);
        descReset.setEnabled(!resetStringEquals(descArea.getText(), currentMedia.getOriginalDescription()));
    }

    private void dateChanged(LocalDate date) {
        currentMedia.setDate(date);
        dateReset.setEnabled(!Objects.equals(date, currentMedia.getOriginalDate()));
    }

    private void timeChanged(LocalTime time) {
        currentMedia.setTime(time);
        timeReset.setEnabled(!Objects.equals(time, currentMedia.getOriginalTime()));
    }

    private static ImageMediaComponent getMissingImageComponent() {
        try {
            BufferedImage bi = UIUtils.loadBufferedImage(ClassLoader.getSystemResource("missing-image-icon.svg"), -1, -1);
            return new ImageMediaComponent(bi);
        } catch (IOException e) {
            ErrorDialog ed = new ErrorDialog("Internal Error", "A fatal error has occured!", e);
            ed.setVisible(true);
            System.exit(0);
            return null;
        }
    }

    private void updateResetButtons() {
        nameReset.setEnabled(!resetStringEquals(nameField.getText(), currentMedia.getOriginalName()));
        descReset.setEnabled(!resetStringEquals(descArea.getText(), currentMedia.getOriginalDescription()));
        dateReset.setEnabled(!Objects.equals(datePicker.getDate(), currentMedia.getOriginalDate()));
        timeReset.setEnabled(!Objects.equals(timePicker.getTime(), currentMedia.getOriginalTime()));
    }

    private static boolean resetStringEquals(String s1, String s2) {
        if (s1 != null && s1.isEmpty()) {
            s1 = null;
        }
        if (s2 != null && s2.isEmpty()) {
            s2 = null;
        }
        return Objects.equals(s1, s2);
    }

    private void mapButtonHandler(ActionEvent e) {
        Window wa = SwingUtilities.getWindowAncestor(this);
        LocationSelectDialog diag = new LocationSelectDialog(wa,
                currentMedia.getLatitude(), currentMedia.getLongitude(), currentMedia.getOriginalLatitude(),
                currentMedia.getOriginalLongitude());
        if (wa != null) {
            diag.setSize((2 * wa.getWidth()) / 3, (2 * wa.getHeight()) / 3);
            UIUtils.centerWindow(diag, wa);
        }
        diag.setVisible(true);
        if (diag.getResponse() == LocationSelectDialog.RESPONSE_SELECT) {
            GeoPosition pos = diag.getCurrentLocation();
            if (pos == null) {
                currentMedia.setLatitude(null);
                currentMedia.setLongitude(null);
            } else {
                currentMedia.setLatitude(pos.getLatitude());
                currentMedia.setLongitude(pos.getLongitude());
            }
            mapValue.setText(formatLocation(currentMedia));
        }
    }

    public void updateName() {
        nameChanged(nameField.getText());
    }
}
