package zander.ui.media.map;

import java.awt.BorderLayout;
import java.awt.Window;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.lang.reflect.InvocationTargetException;
import java.text.DecimalFormat;
import java.util.Objects;
import java.util.concurrent.TimeUnit;

import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.JButton;
import javax.swing.JDialog;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JSpinner;
import javax.swing.JSplitPane;
import javax.swing.JTextField;
import javax.swing.SpinnerNumberModel;
import javax.swing.SwingUtilities;
import javax.swing.event.ChangeListener;
import javax.swing.event.DocumentEvent;
import javax.swing.event.DocumentListener;
import javax.swing.event.MouseInputListener;

import org.jxmapviewer.JXMapViewer;
import org.jxmapviewer.OSMTileFactoryInfo;
import org.jxmapviewer.input.CenterMapListener;
import org.jxmapviewer.input.PanKeyListener;
import org.jxmapviewer.input.PanMouseInputListener;
import org.jxmapviewer.input.ZoomMouseWheelListenerCursor;
import org.jxmapviewer.viewer.DefaultTileFactory;
import org.jxmapviewer.viewer.GeoPosition;
import org.jxmapviewer.viewer.TileFactoryInfo;

import zander.ui.UIUtils;
import zander.ui.docs.DocumentationViewer;

public class LocationSelectDialog extends JDialog {
    private static final long serialVersionUID = 817598179351293L;

    public static final int RESPONSE_CANCEL = 0;
    public static final int RESPONSE_SELECT = 1;

    private static TileFactoryInfo mapTileInfo;
    private static DefaultTileFactory mapTileFactory;
    static {
        final Runnable initAction = () -> {
            mapTileInfo = new OSMTileFactoryInfo();
            mapTileFactory = new DefaultTileFactory(mapTileInfo);
            mapTileFactory.setThreadPoolSize(8);
        };
        try {
            if (SwingUtilities.isEventDispatchThread()) {
                initAction.run();
            } else {
                SwingUtilities.invokeAndWait(initAction);
            }
        } catch (InvocationTargetException | InterruptedException e) {
            throw new Error(e);
        }
    }
    @SuppressWarnings("unused")
    private static final LocationEncoderCacheCleaner locationEncoderCacheCleaner;
    private static final LocationEncoder locationEncoder;
    static {
        locationEncoder = new OSMLocationEncoder();
        locationEncoderCacheCleaner = new LocationEncoderCacheCleaner(locationEncoder, TimeUnit.MINUTES.toMillis(5));
    }

    private final ActionListener searchActionListener = new ActionListener() {
        @Override
        public void actionPerformed(ActionEvent e) {
            if (!searchField.getText().isBlank() && locationSelect.search(searchField.getText())
                && splitPane.getParent() == null) {
                remove(mapComp);
                splitPane.setLeftComponent(locationSelect);
                splitPane.setRightComponent(mapComp);
                add(splitPane, BorderLayout.CENTER);
                revalidate();
                splitPane.setDividerLocation(0.3);
            }
        }
    };
    private final GeoPosition origPos;

    private final JLabel searchLabel;
    private final JTextField searchField;
    private final JButton searchButton;
    private final JPanel searchPanel;
    private final JLabel locationLabel;
    private final LocationSelectPanel locationSelect;
    private final JPanel locationPanel;
    private final MapSelectionPainter mapPainter;
    private final JXMapViewer mapComp;
    private final JSplitPane splitPane;
    private final JButton cancelButton;
    private final JLabel zoomLabel;
    private final SpinnerNumberModel zoomModel;
    private final JSpinner zoomField;
    private final JLabel posLabel;
    private final SpinnerNumberModel latModel;
    private final JSpinner latField;
    private final JLabel posSepLabel;
    private final SpinnerNumberModel lonModel;
    private final JSpinner lonField;
    private final JButton resetButton;
    private final JButton helpButton;
    private final JButton clearButton;
    private final JButton selectButton;
    private final JPanel buttonPanel;
    private final JPanel content;

    private int response = RESPONSE_CANCEL;
    private GeoPosition currPos;
    private boolean disableUpdates = false;

    public LocationSelectDialog(Window parent, Double lat, Double lon, Double origLat, Double origLon) {
        super(parent, "Select Location");
        origPos = makeGeoPosition(origLat, origLon);
        searchLabel = new JLabel("Search:");
        searchField = new JTextField();
        searchField.getDocument().addDocumentListener(new DocumentListener(){
                @Override
                public void insertUpdate(DocumentEvent e) { update(); }
                @Override
                public void removeUpdate(DocumentEvent e) { update(); }
                @Override
                public void changedUpdate(DocumentEvent e) { update(); }
                private void update() {
                    searchButton.setEnabled(!searchField.getText().isBlank());
                }
            });
        searchField.addActionListener(searchActionListener);
        searchButton = new JButton("Search");
        searchButton.setMnemonic(KeyEvent.VK_E);
        searchButton.addActionListener(searchActionListener);
        searchPanel = new JPanel();
        searchPanel.setLayout(new BoxLayout(searchPanel, BoxLayout.X_AXIS));
        searchPanel.add(searchLabel);
        searchPanel.add(searchField);
        searchPanel.add(searchButton);
        locationLabel = new JLabel("Results:");
        locationSelect = new LocationSelectPanel(locationEncoder);
        locationSelect.addLocationSelectListener((e) -> updateSelection(e.getEntry().getPosition(), false, true, true));
        locationPanel = new JPanel(new BorderLayout());
        locationPanel.setVisible(false);
        locationPanel.add(locationLabel, BorderLayout.PAGE_START);
        locationPanel.add(locationSelect, BorderLayout.CENTER);
        mapPainter = new MapSelectionPainter();
        mapComp = new JXMapViewer();
        mapComp.setTileFactory(mapTileFactory);
        mapComp.setOverlayPainter(mapPainter);
        mapComp.addPropertyChangeListener("zoom", new PropertyChangeListener() {
            @Override
            public void propertyChange(PropertyChangeEvent evt) {
                pushUpdate(() -> {
                    zoomModel.setValue(20 - mapComp.getZoom());
                });
            }
        });
        MouseInputListener mapMia = new PanMouseInputListener(mapComp);
        mapComp.addMouseListener(mapMia);
        mapComp.addMouseMotionListener(mapMia);
        mapComp.addMouseListener(new CenterMapListener(mapComp));
        mapComp.addMouseWheelListener(new ZoomMouseWheelListenerCursor(mapComp));
        mapComp.addKeyListener(new PanKeyListener(mapComp));
        mapComp.addMouseListener(new MouseAdapter() {
            @Override
            public void mouseClicked(MouseEvent e) {
                updateSelection(mapComp.convertPointToGeoPosition(e.getPoint()), true, false, true);
            }
        });
        splitPane = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT);
        splitPane.setOneTouchExpandable(true);
        splitPane.setLeftComponent(locationPanel);
        splitPane.setRightComponent(mapComp);
        cancelButton = new JButton("Cancel");
        cancelButton.setMnemonic(KeyEvent.VK_C);
        cancelButton.addActionListener((e) -> dispose());
        zoomLabel = new JLabel("Zoom:");
        zoomModel = new SpinnerNumberModel(13, 1, 20, 1);
        zoomField = new JSpinner(zoomModel);
        zoomField.addChangeListener((e) -> {
                pushUpdate(() -> {
                        mapComp.setZoom(20 - zoomModel.getNumber().intValue());
                    });
            });
        posLabel = new JLabel("Selection:");
        latModel = new SpinnerNumberModel(0.0, -90.0, 90.0, 1.0);
        latField = new JSpinner(latModel);
        final JSpinner.NumberEditor latEditor = (JSpinner.NumberEditor) latField.getEditor();
        final DecimalFormat latFormat = latEditor.getFormat();
        latFormat.setMaximumFractionDigits(4);
        latFormat.setMinimumFractionDigits(0);
        posSepLabel = new JLabel(",");
        lonModel = new SpinnerNumberModel(0.0, -180.0, 180.0, 1.0);
        lonField = new JSpinner(lonModel);
        final JSpinner.NumberEditor lonEditor = (JSpinner.NumberEditor) lonField.getEditor();
        final DecimalFormat lonFormat = lonEditor.getFormat();
        lonFormat.setMaximumFractionDigits(4);
        lonFormat.setMinimumFractionDigits(0);
        final ChangeListener posListener = (e) -> updateSelection(latModel.getNumber().doubleValue(),
                                                                  lonModel.getNumber().doubleValue(), true, true, false);
        latField.addChangeListener(posListener);
        lonField.addChangeListener(posListener);
        resetButton = new JButton("↺");
        helpButton = new JButton("?");
        helpButton.addActionListener((e) -> DocumentationViewer.show(this, "Media/Selecting Locations"));
        clearButton = new JButton("Clear");
        clearButton.setMnemonic(KeyEvent.VK_L);
        clearButton.addActionListener((e) -> {
                int resp = JOptionPane.showConfirmDialog(this, "Are you sure you would like to clear the current selection?",
                    "Confirm Clear", JOptionPane.YES_NO_OPTION, JOptionPane.QUESTION_MESSAGE);
                if (resp == JOptionPane.YES_OPTION) {
                    currPos = null;
                    response = RESPONSE_SELECT;
                    dispose();
                }
            });
        resetButton.addActionListener((e) -> updateSelection(origPos, true, true, true));
        selectButton = new JButton("Select");
        selectButton.setMnemonic(KeyEvent.VK_S);
        selectButton.addActionListener((e) -> {
                response = RESPONSE_SELECT;
                dispose();
            });
        buttonPanel = new JPanel();
        buttonPanel.setLayout(new BoxLayout(buttonPanel, BoxLayout.X_AXIS));
        buttonPanel.add(cancelButton);
        buttonPanel.add(Box.createHorizontalGlue());
        buttonPanel.add(zoomLabel);
        buttonPanel.add(zoomField);
        buttonPanel.add(Box.createHorizontalGlue());
        buttonPanel.add(posLabel);
        buttonPanel.add(latField);
        buttonPanel.add(posSepLabel);
        buttonPanel.add(lonField);
        buttonPanel.add(resetButton);
        buttonPanel.add(Box.createHorizontalGlue());
        buttonPanel.add(helpButton);
        buttonPanel.add(clearButton);
        buttonPanel.add(selectButton);
        content = new JPanel(new BorderLayout());
        content.add(searchPanel, BorderLayout.PAGE_START);
        content.add(mapComp, BorderLayout.CENTER);
        content.add(buttonPanel, BorderLayout.PAGE_END);
        setContentPane(content);
        setResizable(true);
        setAlwaysOnTop(true);
        setModalityType(JDialog.DEFAULT_MODALITY_TYPE);
        setDefaultCloseOperation(JDialog.DISPOSE_ON_CLOSE);
        UIUtils.centerWindow(this, parent);
        updateSelection(lat, lon, false, true, true);
        mapComp.setZoom(7);
        if (lat == null || lon == null) {
            mapPainter.setCurrentPosition(null);
            repaint();
        }
    }

    public int getResponse() {
        return response;
    }

    public GeoPosition getCurrentLocation() {
        return currPos;
    }

    private void updateSelection(Double lat, Double lon, boolean clearListSel, boolean updateMap, boolean updateField) {
        updateSelection(makeGeoPosition(lat, lon), clearListSel, updateMap, updateField);
    }

    private void updateSelection(GeoPosition pos, boolean clearListSel, boolean updateMap, boolean updateField) {
        currPos = pos;
        updateResetButton();
        pushUpdate(() -> {
            if (clearListSel) {
                locationSelect.clearSelection();
            }
            if (updateMap) {
                mapComp.setAddressLocation(pos);
            }
            if (updateField) {
                latModel.setValue(pos.getLatitude());
                lonModel.setValue(pos.getLongitude());
            }
        });
        mapPainter.setCurrentPosition(pos);
        repaint();
    }

    private void updateResetButton() {
        resetButton.setEnabled(!Objects.equals(currPos, origPos));
    }

    private GeoPosition makeGeoPosition(Double lat, Double lon) {
        if (lat == null || lon == null) {
            return new GeoPosition(0.0, 0.0);
        }
        return new GeoPosition(lat, lon);
    }

    private void pushUpdate(Runnable update) {
        if (!disableUpdates) {
            disableUpdates = true;
            update.run();
            disableUpdates = false;
        }
    }
}
