package zander.ui.transferhandlers;

import java.awt.Component;
import java.awt.datatransfer.DataFlavor;
import java.awt.datatransfer.Transferable;
import java.awt.datatransfer.UnsupportedFlavorException;
import java.awt.dnd.DragSource;
import java.io.IOException;
import java.util.Objects;

import javax.swing.DefaultListModel;
import javax.swing.JComponent;
import javax.swing.JList;
import javax.swing.TransferHandler;

/** From this StackOverflow post: {@link https://stackoverflow.com/questions/16586562/reordering-jlist-with-drag-and-drop} */
public class ListArrangeTransferHandler extends TransferHandler {
    private static final long serialVersionUID = 1L;

    private final DataFlavor objectArrayDataFlavor = new DataFlavor(Object[].class, "An object");

    private int[] indices;
    private int addIndex = -1;
    private int addCount;

    private final Runnable before;
    private final Runnable after;

    public ListArrangeTransferHandler() {
        this(null, null);
    }

    public ListArrangeTransferHandler(Runnable before, Runnable after) {
        this.before = before;
        this.after = after;
    }

    @Override
    protected Transferable createTransferable(JComponent c) {
        if (before != null) {
            before.run();
        }
        JList<?> source = (JList<?>) c;
        c.getRootPane().getGlassPane().setVisible(true);
        indices = source.getSelectedIndices();
        Object[] transferObjects = source.getSelectedValuesList().toArray();
        return new Transferable() {
            @Override
            public DataFlavor[] getTransferDataFlavors() {
                return new DataFlavor[] { objectArrayDataFlavor };
            }

            @Override
            public boolean isDataFlavorSupported(DataFlavor flavor) {
                return Objects.equals(objectArrayDataFlavor, flavor);
            }

            @Override
            public Object getTransferData(DataFlavor flavor) throws UnsupportedFlavorException, IOException {
                if (!isDataFlavorSupported(flavor)) {
                    throw new UnsupportedFlavorException(flavor);
                }
                return transferObjects;
            }
        };
    }

    @Override
    public boolean canImport(TransferSupport info) {
        return info.isDrop() && info.isDataFlavorSupported(objectArrayDataFlavor);
    }

    @Override
    public int getSourceActions(JComponent c) {
        Component glassPane = c.getRootPane().getGlassPane();
        glassPane.setCursor(DragSource.DefaultMoveDrop);
        return MOVE;
    }

    @SuppressWarnings({"unchecked", "rawtypes"})
    @Override
    public boolean importData(TransferSupport info) {
        TransferHandler.DropLocation tdl = info.getDropLocation();
        if (!canImport(info) || !(tdl instanceof JList.DropLocation)) {
            return false;
        }
        JList.DropLocation dl = (JList.DropLocation) tdl;
        JList target = (JList) info.getComponent();
        DefaultListModel listModel = (DefaultListModel) target.getModel();
        int max = listModel.getSize();
        int index = dl.getIndex();
        index = index < 0 ? max : index;
        index = Math.min(index, max);
        addIndex = index;
        try {
            Object[] values = (Object[]) info.getTransferable().getTransferData(objectArrayDataFlavor);
            for (int i = 0; i < values.length; ++i) {
                int idx = index++;
                listModel.add(idx, values[i]);
                target.addSelectionInterval(idx, idx);
            }
            addCount = values.length;
            return true;
        } catch (UnsupportedFlavorException | IOException e) {
            throw new Error("Flavor was invalid or an IOException occured", e);
        }
    }

    @Override
    protected void exportDone(JComponent c, Transferable data, int action) {
        c.getRootPane().getGlassPane().setVisible(false);
        cleanup(c, action == MOVE);
    }

    @SuppressWarnings("rawtypes")
    private void cleanup(JComponent c, boolean remove) {
        if (remove && Objects.nonNull(c)) {
            if (addCount > 0) {
                for (int i = 0; i < indices.length; ++i) {
                    if (indices[i] >= addIndex) {
                        indices[i] += addCount;
                    }
                }
            }
            JList source = (JList) c;
            DefaultListModel model = (DefaultListModel) source.getModel();
            for (int i = indices.length - 1; i >= 0; --i) {
                model.remove(indices[i]);
            }
        }
        indices = null;
        addCount = 0;
        addIndex = -1;
        if (after != null) {
            after.run();
        }
    }

}
